<?php
namespace AIOSEO\Plugin\Addon\Redirects\Main;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use AIOSEO\Plugin\Addon\Redirects\Utils;
use AIOSEO\Plugin\Addon\Redirects\Models;

/**
 * Main class to run our 404 redirects.
 *
 * @since 1.2.2
 */
class Redirect404 {
	/**
	 * Class constructor.
	 *
	 * @since 1.2.2
	 */
	public function __construct() {
		// Maybe redirect 404s.
		add_action( 'template_redirect', [ $this, 'templateRedirect' ], 2000 );

		// Maybe short-circuit WP's redirect guess.
		add_filter( 'pre_redirect_guess_404_permalink', [ $this, 'redirectGuess' ] );
	}

	/**
	 * Short-circuit WP's redirect guess.
	 *
	 * @since 1.2.2
	 *
	 * @param  string|null $guessUrl The pre guess url.
	 * @return void
	 */
	public function redirectGuess( $guessUrl ) {
		if ( ! Utils\Request::isRedirectTest() ) {
			$this->maybeRedirect();
		}

		return $guessUrl;
	}

	/**
	 * Intercept the request and redirect after logs have been saved.
	 *
	 * @since 1.2.2
	 *
	 * @return void
	 */
	public function templateRedirect() {
		// Skip redirecting if it's a redirect test.
		if ( ! is_404() || Utils\Request::isRedirectTest() ) {
			return;
		}

		$this->maybeRedirect();
	}

	/**
	 * Tries to redirect a 404.
	 *
	 * @since 1.2.2
	 *
	 * @return void
	 */
	public function maybeRedirect() {
		// We'll try redirecting to the parent first.
		if ( aioseoRedirects()->options->advanced404s->redirectToParent ) {
			$this->parentRedirect();
		}

		// Last fallback is to the home page.
		if ( aioseoRedirects()->options->advanced404s->redirectToHome ) {
			aioseoRedirects()->helpers->do404Redirect( get_home_url(), 'HOME' );
		}
	}

	/**
	 * Tries to redirect to a stored parent reference.
	 *
	 * @since 1.2.2
	 *
	 * @return void
	 */
	private function parentRedirect() {
		if ( ! aioseo()->license->hasAddonFeature( 'aioseo-redirects', '404-parent-redirect' ) ) {
			return;
		}

		$requestUrl  = Utils\Request::getRequestUrl();
		$redirect404 = Models\Redirect404::getRedirectByUrl( $requestUrl );
		if ( ! $redirect404->exists() ) {
			return;
		}

		$redirectParent = new RedirectParent404( $redirect404 );

		// Redirects below are in order of importance.
		// 1. First the parent post.
		$redirectParent->postParentRedirect();

		// 2. Second the parent term.
		$redirectParent->termParentRedirect();

		// 3. WooCommerce redirect to the shop page.
		$redirectParent->woocommerceParentRedirect();

		// 4. Lastly, let's try the post type archive.
		$redirectParent->postTypeArchiveRedirect();
	}
}